/* This file is part of turtle, a turtle-graphics system for Scheme.

Copyright (C) 2008  Luca Saiu

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#include "turtle.h"
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <math.h>

/* Allocate a new turtle, initialize it and return it: */
turtle_turtle_t turtle_make_turtle(void){
  turtle_turtle_t result = (turtle_turtle_t)
    turtle_malloc(sizeof(struct turtle_turtle));
  turtle_initialize_turtle(result);
  return result;
}

/* Destroy a turtle: */
void turtle_destroy_turtle(turtle_turtle_t turtle){
  turtle_free(turtle);
}

void turtle_set_pen(turtle_turtle_t turtle, const bool down){
  turtle->pen_down = down;
}

void turtle_set_color(turtle_turtle_t turtle, const turtle_color_t color){
  /* Just copy the color into the appropriate field: */
  memcpy(&(turtle->color), color, sizeof(struct turtle_color));
}

void turtle_set_cursor_height(turtle_turtle_t turtle, const float height){
  turtle->cursor_height = height;
}

void turtle_forward(turtle_turtle_t turtle, const float distance){
  /* Compute the new position (note that the orientation does not
     change): */
  const float new_x = turtle->x + cosf(turtle->orientation) * distance;
  const float new_y = turtle->y + sinf(turtle->orientation) * distance;
  
  /* Draw a line if the pen is down: */
  if(turtle->pen_down)
    turtle_draw_line(turtle->x, turtle->y, new_x, new_y, & turtle->color);
  
  /* Update the position: */
  turtle->x = new_x;
  turtle->y = new_y;
}

void turtle_backward(turtle_turtle_t turtle, const float distance){
  /* This is just like 'forward' called with a distance of the 
     opposite sign: */
  turtle_forward(turtle, - distance);
}

void turtle_left_radiants(turtle_turtle_t turtle, const float angle){
  /* Just change the angle, without drawing anything: */
  turtle->orientation -= angle;
}

void turtle_right_radiants(turtle_turtle_t turtle, const float angle){
  /* This is just like 'left' called with an angle of the 
     opposite sign: */
  turtle_left_radiants(turtle, - angle);
}

static float degrees_to_radiants(const float degrees){
  return degrees / 180.0 * M_PI;
}

static float radiants_to_degrees(const float radiants){
  return radiants / M_PI * 180.0;
}

void turtle_left_degrees(turtle_turtle_t turtle, const float angle){
  turtle_left_radiants(turtle, degrees_to_radiants(angle));
}

void turtle_right_degrees(turtle_turtle_t turtle, const float angle){
  turtle_right_radiants(turtle, degrees_to_radiants(angle));
}

void turtle_set_orientation_radiants(turtle_turtle_t turtle, const float new_orientation){
  turtle->orientation = - M_PI / 2.0 + new_orientation;
}

void turtle_set_orientation_degrees(turtle_turtle_t turtle, const float new_orientation){
  turtle_set_orientation_radiants(turtle, degrees_to_radiants(new_orientation));
}

void turtle_set_position(turtle_turtle_t turtle, const float new_x, const float new_y){
  turtle->x = new_x;
  turtle->y = new_y;
}

void turtle_goto(turtle_turtle_t turtle, const float new_x, const float new_y){
  /* Draw a line if the pen is down: */
  if(turtle->pen_down)
    turtle_draw_line(turtle->x, turtle->y, new_x, new_y, & turtle->color);
  
  /* Update the position: */
  turtle->x = new_x;
  turtle->y = new_y;
}

void turtle_home(turtle_turtle_t turtle){
  turtle_set_position(turtle,
                      turtle_get_window_width() / 2,
                      turtle_get_window_height() / 2);
  turtle_set_orientation_radiants(turtle, 0); // up
}

void turtle_hide(turtle_turtle_t turtle){
  turtle->hidden = true;
}

void turtle_show(turtle_turtle_t turtle){
  turtle->hidden = false;
}

/* Initialize a given, *already allocated*, turtle: */
void turtle_initialize_turtle(turtle_turtle_t turtle){
  turtle_home(turtle);
  turtle_set_color(turtle, turtle_green);
  turtle_set_pen(turtle, true);
  turtle_set_cursor_height(turtle, TURTLE_CURSOR_HEIGHT);
  turtle_show(turtle);
}

void turtle_draw_cursor_triangle(const turtle_turtle_t turtle,
                                 const float cursor_height,
                                 const turtle_color_t cursor_color){
  /* Compute the position of each vertex: */
  const float cursor1_x = turtle->x + cosf(turtle->orientation) * cursor_height;
  const float cursor1_y = turtle->y + sinf(turtle->orientation) * cursor_height;

  const float cursor2_x = turtle->x + cosf(turtle->orientation + M_PI / 2) * cursor_height;
  const float cursor2_y = turtle->y + sinf(turtle->orientation + M_PI / 2) * cursor_height;

  const float cursor3_x = turtle->x + cosf(turtle->orientation - M_PI / 2) * cursor_height;
  const float cursor3_y = turtle->y + sinf(turtle->orientation - M_PI / 2) * cursor_height;
  
  /* Draw a triangle: */
  turtle_draw_cursor_line(cursor1_x, cursor1_y, cursor2_x, cursor2_y, cursor_color);
  turtle_draw_cursor_line(cursor1_x, cursor1_y, cursor3_x, cursor3_y, cursor_color);
  turtle_draw_cursor_line(cursor2_x, cursor2_y, cursor3_x, cursor3_y, cursor_color);
}

void turtle_draw_cursor(turtle_turtle_t turtle){
  /* Do nothing if the turtle is hidden: */
  if(turtle->hidden)
    return;
  
  /* The cursor is made of two triangles: the innder one is only shown if the pen is
     down: */
  turtle_draw_cursor_triangle(turtle, turtle->cursor_height, turtle_green);
  if(turtle->pen_down)
    turtle_draw_cursor_triangle(turtle, (int)(turtle->cursor_height * 0.8), turtle_green);
}
